#include "WGraph.h"

WGraph::WGraph() {
    m_size = 0;
    m_adj = NULL;
    m_conn = NULL;
}

WGraph::WGraph(unsigned int sz) {
    m_size = sz;
    //allocate sz * sz adj matrix
    m_adj = new double* [sz];
    m_conn = new double* [sz];
    for (int i = 0; i < m_size; ++i) {
        m_adj[i] = new double[sz];
        m_conn[i] = new double[sz];
    }
    //start with edges
    for (int i = 0; i < m_size; ++i) {
        for (int j = 0; j < m_size; ++j) {
            m_adj[i][j] = std::numeric_limits<double>::max(); //essentially infinity
            m_conn[i][j] = std::numeric_limits<double>::max();
        }
    }
}

WGraph::~WGraph() {
    //release memory
}

void WGraph::addEdge(VertexID i, VertexID j, double w) {
    if (i < m_size && j < m_size) {
        m_adj[i][j] = w;
        m_adj[j][i] = w;
    }
}

void WGraph::removeEdge(VertexID i, VertexID j) {
    if (i < m_size && j < m_size) {
        m_adj[i][j] = std::numeric_limits<double>::max();
        m_adj[j][i] = std::numeric_limits<double>::max();
    }
}

bool WGraph::areAdjacent(VertexID i, VertexID j) {
    return (m_adj[i][j] < std::numeric_limits<double>::max());
}


void WGraph::calcFW() { //runtime complexity O(v^3)
    for (int i = 0; i < m_size; ++i) {
        for (int j = 0; j < m_size; ++j) {
            m_conn[i][j] = m_adj[i][j]; //start with conn == adj matrix
        }
    }
    for (int im = 0; im < m_size; ++im) { //intermediate points --> transitive closure
        for (int source = 0; source < m_size; ++source) { //source = starting point
            for (int sink = 0; sink < m_size; ++sink) { //sink = ending point
                if (source == sink) {
                    continue;
                }
                else if (m_conn[source][im] != std::numeric_limits<double>::max() &&
                    m_conn[im][sink] != std::numeric_limits<double>::max() &&
                    m_conn[source][sink] > m_conn[source][im] + m_conn[im][sink]) {
                    m_conn[source][sink] = m_conn[source][im] + m_conn[im][sink];
                }
            }
        }
    }
}

double WGraph::cheapestCost(VertexID i, VertexID j) {
    return m_conn[i][j]; //constant
}

int WGraph::parentOf(int source)
{
    if (m_parent[source] != std::numeric_limits<unsigned int>::max())
    {
        m_parent[source] = parentOf(m_parent[source]);
        return m_parent[source];
    }
    return source;
}

void WGraph::unionSet(int source, int sink)
{
    int source_parent = parentOf(source);
    int sink_parent = parentOf(sink);

    if (source_parent != sink_parent) {
        if (m_rank[source_parent] < m_rank[sink_parent]) {
            m_parent[source_parent] = sink_parent;
        }
        else if (m_rank[source_parent] > m_rank[sink_parent]) {
            m_parent[sink_parent] = source_parent;
        }
        else {
            m_parent[sink_parent] = source_parent;
            m_rank[source_parent] += 1;
        }
    }
}

void WGraph::sortEdges(double** m_mst_adj, unsigned int* sources, unsigned int* sinks, int& numOfEdges) {

    numOfEdges = 0;

    for (int i = 0; i < m_size; ++i) {
        for (int j = 0; j < m_size; ++j) {
            if (m_mst_adj[i][j] != 0) {
                sources[numOfEdges] = i;
                sinks[numOfEdges] = j;

                numOfEdges++;
            }
            m_mst_adj[i][j] = 0;
            m_mst_adj[j][i] = 0;
        }
    }

    for (int i = 0; i < numOfEdges - 1; ++i) {
        for (int j = 0; j < numOfEdges - i - 1; ++j) {
            if (m_adj[sources[j]][sinks[j]] < m_adj[sources[j + 1]][sinks[j + 1]])
            {
                unsigned int temp = sources[j];
                sources[j] = sources[j + 1];
                sources[j + 1] = temp;

                temp = sinks[j];
                sinks[j] = sinks[j + 1];
                sinks[j + 1] = temp;
            }
        }
    }
}

void WGraph::computeMST() {

    int totalCost = 0;

    double** m_mst_adj;

    m_mst_adj = new double* [m_size];
    for (int i = 0; i < m_size; ++i) {
        m_mst_adj[i] = new double[m_size];
        for (int j = 0; j < m_size; ++j)
            m_mst_adj[i][j] = areAdjacent(i, j) ? m_adj[i][j] : 0;
    }

    unsigned int* sources, *sinks;
    sources = new unsigned int[m_size * m_size];
    sinks = new unsigned int[m_size * m_size];
    int numOfEdges = 0;
    sortEdges(m_mst_adj, sources, sinks, numOfEdges);

    for (int i = 0; i < numOfEdges; ++i) {
        std::cout << sources[i] << " " << sinks[i] << " = " << m_adj[sources[i]][sinks[i]] << std::endl;
    }

    m_parent = new unsigned int[m_size];
    m_rank = new unsigned int[m_size];

    for (int source = 0; source < m_size; ++source) {
        m_parent[source] = std::numeric_limits<unsigned int>::max();
        m_rank[source] = 1;
    }

    for (int i = 0; i < numOfEdges; ++i) {

        int source = sources[i], sink = sinks[i];

        if (areAdjacent(source, sink) && (parentOf(source) != parentOf(sink))) {

            int weight = m_adj[source][sink];

            unionSet(source, sink);

            m_mst_adj[source][sink] = weight;

            totalCost += weight;

            break;
        }
        
    }
    
    std::cout << "Cost of the MST: " << totalCost << std::endl << std::endl;

    std::cout << "Adjacency matrix representation of the MST: " << std::endl;
    for (int i = 0; i < m_size; ++i) {
        for (int j = 0; j < m_size; ++j) {
            std::cout << m_mst_adj[i][j] << " ";
        }
        std::cout << std::endl;
    }   
}